from RoboidControl.Thing import Thing

from typing import Union, Optional

class Participant:
    """! A participant is a device which manages things.

    It can communicate with other participant to synchronise the state of things.
    This class is used to register the things the participant is managing.
    It also maintains the communcation information to contact the participant.
    It is used as a basis for the local participant, but also as a reference to remote participants.
    """

    def __init__(self, ip_address: str, port: int) -> None:
        """! Create a new participant with the given communcation info
        @param ip_address The IP address of the participant
        @param port The UDP port of the participant
        """

        ## The Ip Address of a participant. When the participant is local, this contains 0.0.0.0
        self.ip_address: str = ip_address
        ## The port number for UDP communication with the participant. This is 0 for isolated participants.
        self.port: int = port

        ## he network Id to identify the participant.
        self.network_id: int = 0

        ## The things managed by this participant
        self.things: set[Thing] = set()

    def Get(self, thing_id: int) -> Optional[Thing]:
        """!  Get the thing with the given properties

        @param thing_id The ID of the thing
        @return The thing if found, None in other cases
        """
        for thing in self.things:
            if thing.id == thing_id:
                return thing
        return None

    def Add(self, thing: Thing, check_id: bool = True):
        """! Add a new thing for this participant. 
        @param thing The thing to add
        @param check_id If true, the thing.id is regenerated if it is zero
        """
        if check_id and thing.id == 0:
            thing.id = len(self.things) + 1
            self.things.add(thing)
        else:
            found_thing = self.Get(thing.id)
            if found_thing == None:
                self.things.add(thing)
    
    def Remove(self, thing: Thing) -> None:
        """! Remove a thing for this participant
        @param thing The thing to remove
        """
        self.things.remove(thing)

    def Update(self, currentTimeMs: int = 0) -> None:
        """! Update all things for this participant
        @param The current time in milliseconds (optional)
        """
        for thing in list(self.things):
            thing.Update(currentTimeMs)

    participants: set['Participant'] = set()

    @staticmethod
    def GetParticipant(arg1: Union[str, int], port: Optional[int] = None) -> Optional['Participant']:
        """! Retrieve a participant
        @param arg1 Either:
        - str: The IP address of the participant
        - int: The network ID of the participant
        @param port The port number of the participant 
        """
        if port is not None:
            ip_address = str(arg1)
            for participant in Participant.participants:
                if participant.ip_address == ip_address and participant.port == port:
                    return participant
            return None
        else:
            network_id = int(arg1)
            for participant in Participant.participants:
                if participant.network_id == network_id:
                    return participant
            return None

    def AddParticipant(arg1: Union[str, 'Participant'], port: Optional[int] = None) -> 'Participant':
        """! Add a participant
        @param arg1 Either:
        - str: The IP address of the participant
        - Participant:  The participant to add
        @param port The port number of the participant 
        """
        if port is not None:
            ip_address = str(arg1)
            participant = Participant(ip_address, port)
            participant.network_id = len(Participant.participants) + 1
            Participant.AddParticipant(participant)
            return participant
        else:
            participant: Participant = arg1
            foundParticipant = Participant.GetParticipant(participant.network_id, participant.port)
            if foundParticipant is not None:
                Participant.participants.add(participant)
            return participant
